/*
 *  linux/arch/i386/kernel/fast_timer.c
 *
 * Imitation being the sincerest form of flattery, this code is
 * patterned after (and reuses data structures from) the ordinary
 * Linux timer code.  The difference here is that times are kept
 * in mini-jiffies (currently 131.6 usec.)
 *
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/mc146818rtc.h>
#include <linux/timex.h>
#include <linux/config.h>

#ifdef BREAK_DEBUG
#define BREAKPOINT() asm("   int $3");
#else
#define BREAKPOINT()
#endif

#ifdef CONFIG_FAST_TIMER
#include "linux/fast_timer.h"

#define MAX_TIME 2147483647	/* 2^31-1*/
static struct fast_timer_list fast_timer_head = {
	&fast_timer_head,
	&fast_timer_head,
	MAX_TIME,
	0,
	NULL
};

static struct fast_timer_list *hold_queue;

static unsigned long mini_jiffies;

#define	HOLDING		2
#define EMPTYING	1
#define EMPTY		0
static int hold_queue_state;

/* Trim a timer from the list and run it */
static inline void run_fast_timer(struct fast_timer_list *timer)
{
	void (*fn)(struct fast_timer_list *);

	if (!timer) return;

	fn = timer->function;

	timer->next->prev = timer->prev;
	timer->prev->next = timer->next;
	timer->next = timer->prev = NULL;
	/*sti();*/
	/* We give the timer function a pointer to its structure.  This
	 * allows it to look at anything it wants; the drawback is that
	 * it is responsible for freeing the structure, etc.
	 */
 	if (fn) {
		fn(timer);
	}
	/*cli();*/
}

/* Run all timers which have expired */
static inline void run_fast_timer_list(unsigned long minijiffies)
{
	struct fast_timer_list *timer;

	hold_queue_state = HOLDING;
	while (((timer = fast_timer_head.next) != &fast_timer_head) && timer->expires <= minijiffies) {
		run_fast_timer(timer);
	}
	hold_queue_state = EMPTYING;
	while (hold_queue) {
BREAKPOINT();
		timer = hold_queue;
		hold_queue = hold_queue->next;
		(void) add_fast_timer(timer);
	}
	hold_queue_state = EMPTY;
}

void do_fast_timer(struct pt_regs *regs)
{
	unsigned long flags;

	save_flags(flags);
	++mini_jiffies;
	run_fast_timer_list(mini_jiffies);

	/* think about user mode, lost tick problems */

	restore_flags(flags);
	return;
}

int add_fast_timer(struct fast_timer_list *timer)
{
	struct fast_timer_list *p;
	unsigned long flags;

	/* Check if too much time is requested */
	if (timer->expires > MAX_TIME)
		return -EINVAL;
	/* We put in a relative, rather than absolute time */
	timer->expires += mini_jiffies;
	save_flags(flags);
	cli();
	if (hold_queue_state == HOLDING) {
BREAKPOINT();
		timer->next = hold_queue;
		hold_queue = timer;
		restore_flags(flags);
		return 0;
	}

	/* Check whether the clock is in danger of wrapping */
	if (hold_queue_state == EMPTY && timer->expires > MAX_TIME) {
		p = &fast_timer_head;
		/* Be careful: check for any expired timers and run them */
		run_fast_timer_list(mini_jiffies);
		/* Now adjust times back */
BREAKPOINT();
		for (p = p->next; p != &fast_timer_head; p = p->next) {
			p->expires -= mini_jiffies;
			p = p->next;
		}
		timer->expires -= mini_jiffies;
		mini_jiffies = 0;
	}

	/* Find where to put in list */
	p = &fast_timer_head;
	do {
		p = p->next;
	} while (timer->expires > p->expires);

	/* Add it in */
	timer->next = p;
	timer->prev = p->prev;
	p->prev = timer;
	timer->prev->next = timer;

	restore_flags(flags);
	return 0;
}

/* Hmmm -- could actually just call del_timer for this one! */
int del_fast_timer(struct fast_timer_list *timer)
{
	int ret = 0;

	if (timer->next) {
		unsigned long flags;
		struct fast_timer_list *next;

		save_flags(flags);
		cli();
		if ((next = timer->next) != NULL) {
			(next->prev = timer->prev)->next = next;
			timer->next = timer->prev = NULL;
			ret = 1;
		}
		restore_flags(flags);
	}
	return ret;
}
#endif
